import argparse
import json
import logging
import sys
from pathlib import Path

from pydantic import ValidationError
from yaml import YAMLError

from .config import ActionConfig
from .manager import SchedremManager
from .util import Messenger, error_message, get_config_file, take_action


def main() -> None:
    parser = argparse.ArgumentParser(description="Task manager and reminder")
    parser.add_argument(
        "--config",
        type=str,
        help="Path to config file for schedules",
    )
    parser.add_argument(
        "--debug",
        action=argparse.BooleanOptionalAction,
        help="Debug mode",
    )
    parser.add_argument(
        "--action",
        type=str,
        help="JSON string to be executed as an action",
    )
    args = parser.parse_args()

    logging.basicConfig(
        level=logging.DEBUG if args.debug else logging.WARNING,
        format="%(message)s",
    )
    logging.getLogger("asyncio").setLevel(logging.WARNING)
    logging.getLogger("watchdog").setLevel(logging.WARNING)

    if args.action:
        try:
            action = ActionConfig(**json.loads(args.action))
            take_action(action)
        except Exception as e:
            msg = f"{e.__class__.__name__}, {e}"
            m = Messenger()
            m.warning(msg)
        sys.exit()

    if args.config:
        yaml_path = Path(args.config).resolve()
    else:
        try:
            yaml_path = get_config_file()
        except Exception as e:
            msg = f"{e.__class__.__name__}, {e}\nProgram exits."
            m = Messenger()
            m.error(msg)
            sys.exit(msg)

    logging.debug("config path: %s\n", yaml_path)

    while True:
        try:
            schedrem = SchedremManager(yaml_path)
            schedrem.run()
        except ValidationError as e:
            logging.debug(e.json(indent=2))
            m = Messenger()
            m.warning(error_message(e.errors()))
        except YAMLError as e:
            msg = f"YAMLError, {e}"
            logging.debug(msg)
            m = Messenger()
            m.warning(msg)
        except FileNotFoundError as e:
            msg = (
                f"{e.__class__.__name__}, "
                f'Config file "{yaml_path}" not found.\nProgram exits.'
            )
            m = Messenger()
            m.error(msg)
            sys.exit(msg)


if __name__ == "__main__":
    main()



import asyncio
import logging
import subprocess
from collections.abc import Callable
from datetime import datetime, timedelta
from pathlib import Path
from typing import cast

import psutil
from watchdog.events import FileSystemEvent, FileSystemEventHandler
from watchdog.observers import Observer
from yaml import safe_load

from .config import SchedremConfig, ScheduleConfig, TimeConfig


class SchedremManager:
    def __init__(self, yaml_path: Path) -> None:
        self.coros: list = []
        self.tasks: list = []
        self.observer = Observer()
        self.observer.schedule(
            SchedremEventHandler(yaml_path, self),
            yaml_path.resolve().parent,
            recursive=False,
        )
        self.config = SchedremConfig(**self.load_yaml(yaml_path))
        if self.config.disabled:
            logging.debug("Schedrem is disabled.\n")
            # kill all existing action processes when disabled
            for process in psutil.process_iter(["name", "cmdline"]):
                try:
                    name = process.info["name"]
                    cmdline = process.info["cmdline"]
                    if name == "schedrem" and cmdline and "--action" in cmdline:
                        process.terminate()
                except Exception as e:
                    logging.debug("Failed to kill process: %s", e)

        else:
            self.set_schedules()

    def load_yaml(self, yaml_path: Path) -> dict:
        with yaml_path.open() as file:
            return safe_load(file) or {}
        return {}

    def week_num(self, weekday: str | list[str] | None) -> int | list[int] | None:
        """According to date.weekday(), monday is 0 and sunday is 6."""
        if type(weekday) is list:
            return cast(list[int], [self.week_num(w) for w in weekday])
        for aweek in self.config.weekdaynames:
            for i, name in enumerate(aweek):
                if weekday == name:
                    return i
        return None

    def set_schedules(self) -> None:
        self.sched_mans = [
            ScheduleManager(sch, self.week_num, self.config.font)
            for sch in self.config.schedules
        ]
        self.coros = [sched.standby for sched in self.sched_mans]

    async def cancel_awaiter(self) -> None:
        """Just a sleep loop waiting for cancellation."""
        while True:
            try:
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                break

    def run(self) -> None:
        logging.debug("weekdaynames: %s\n", self.config.weekdaynames)

        async def crate_tasks() -> None:
            async with asyncio.TaskGroup() as tg:
                self.tasks = [tg.create_task(coro()) for coro in self.coros]
                self.tasks.append(tg.create_task(self.cancel_awaiter()))

        self.observer.start()
        asyncio.run(crate_tasks())
        logging.debug("Tasks have been cancelled!\n")
        self.observer.stop()
        self.observer.join()


class ScheduleManager:
    def __init__(
        self,
        config: ScheduleConfig,
        week_num: Callable,
        font: str | None,
    ) -> None:
        self.sconf = config
        self.sconf.font = self.sconf.font or font
        self.week_num = week_num

    async def standby(self) -> None:
        while True:
            """No time key should be set in a schedule
            if it should be triggered every minute.
            It means the same as * * * * * in crontab.
            """
            if not self.sconf.enabled:
                return
            nexttime = self.nearest_future(self.sconf.time)
            if not nexttime:
                return
            waittime = (nexttime - datetime.now()).total_seconds()
            logging.debug(self.sconf)
            logging.debug(nexttime)
            logging.debug("%s seconds left\n", waittime)

            try:
                await asyncio.sleep(waittime)
                while datetime.now() < nexttime:
                    """Adjust the gap between monotonic and real-time clocks
                    if sleep time is not enough for nexttime.
                    """
                    await asyncio.sleep(0.1)
                action_arg = self.sconf.model_dump_json(
                    include={
                        "yesno",
                        "command",
                        "message",
                        "sound",
                        "font",
                    },
                )
                subprocess.Popen(["schedrem", "--action", action_arg])
            except asyncio.CancelledError:
                break

    def nearest_future(self, tconf: TimeConfig) -> datetime | None:
        now = datetime.now().replace(second=0, microsecond=0)
        now += timedelta(minutes=1)  # trigger from next minute

        if self.sconf.wait:
            candidate = datetime(**self.sconf.wait.model_dump())
            candidate = max(candidate, now)
        else:
            candidate = now

        if type(tconf.year) is list:
            if candidate.year > max(tconf.year):
                return None
            if candidate.year not in tconf.year:
                candidate = datetime(
                    min([y for y in tconf.year if y > candidate.year]),
                    min(tconf.month) if type(tconf.month) is list else 1,
                    min(tconf.day) if type(tconf.day) is list else 1,
                )

        if (
            type(tconf.month) is list
            and candidate.month not in tconf.month
            or type(tconf.day) is list
            and candidate.day not in tconf.day
        ):
            delta = {"days": 1}
            candidate = datetime(
                candidate.year,
                candidate.month,
                candidate.day,
                min(tconf.hour) if type(tconf.hour) is list else 0,
                min(tconf.minute) if type(tconf.minute) is list else 0,
            )
        else:
            delta = {"minutes": 1}

        def time_match(candi: int, setting: list[int] | None) -> bool:
            return setting is None or type(setting) is list and candi in setting

        while tconf.year is None or (
            type(tconf.year) is list
            and min(tconf.year) <= candidate.year <= max(tconf.year)
        ):
            if (
                time_match(candidate.year, tconf.year)
                and time_match(candidate.month, tconf.month)
                and time_match(candidate.day, tconf.day)
                and time_match(candidate.hour, tconf.hour)
                and time_match(candidate.minute, tconf.minute)
                and time_match(candidate.weekday(), self.week_num(tconf.weekday))
            ):
                return candidate
            candidate += timedelta(**delta)

        return None


class SchedremEventHandler(FileSystemEventHandler):
    def __init__(self, yaml_path: Path, app_manager: SchedremManager) -> None:
        self.yaml_path = yaml_path
        self.app_manager = app_manager

    def is_yaml_path(self, path: str) -> bool:
        return Path(path) == self.yaml_path

    def on_any_event(self, event: FileSystemEvent) -> None:
        if event.event_type in [
            "modified",
            "moved",
            "deleted",
        ] and self.is_yaml_path(event.src_path):
            for task in self.app_manager.tasks:
                task.cancel()



import contextlib
import logging
import os
import subprocess
import sys
import threading
import wave
from pathlib import Path
from tkinter import TclError, Tk, messagebox

import pyaudio

from .config import ActionConfig


def program_dir() -> Path:
    if getattr(sys, "frozen", False):
        return Path(sys.executable).resolve().parent
    return Path(__file__).resolve().parent


def take_action(action: ActionConfig) -> None:
    if action.command and not action.yesno:
        subprocess.Popen(action.command, shell=True)
    yes = False
    if action.yesno:
        logging.debug("yesno: %s\n", action.yesno)
        m = Messenger(action.sound, action.font)
        yes = m.yesno(action.yesno)
    if action.command and yes:
        subprocess.Popen(action.command, shell=True)
    if action.message:
        logging.debug("message: %s\n", action.message)
        m = Messenger(action.sound, action.font)
        m.message(action.message)


class Messenger:
    prg_dir: Path = program_dir()
    default_sound: Path = prg_dir / "assets/alert.wav"
    icon: Path = prg_dir / "assets/schedrem.ico"

    def __init__(
        self,
        sound: str | bool | None = None,
        font: str | None = None,
    ) -> None:
        self.sound_path: Path | None
        if sound is True:
            self.sound_path = self.default_sound
        elif type(sound) is str:
            self.sound_path = Path(os.path.expandvars(sound)).expanduser()
            if not self.sound_path.exists():
                self.sound_path = self.default_sound
        else:
            self.sound_path = None

        self.font = font
        self.keep: bool = True

    def prepare_root(self) -> None:
        self.root = Tk()
        with contextlib.suppress(TclError):
            self.root.iconbitmap(str(self.icon))
        # make Tk blank window invisible
        self.root.withdraw()
        self.root.attributes("-topmost", True)
        self.root.option_add("*Dialog.msg.font", self.font or "Arial 19")
        self.root.option_add("*Dialog.msg.wrapLength", "800p")

    def sing(self) -> None:
        if self.sound_path is None:
            return
        self.keep = True
        with wave.open(str(self.sound_path), "rb") as wf:
            sampwidth = wf.getsampwidth()
            framerate = wf.getframerate()
            channels = wf.getnchannels()
            frames = wf.readframes(wf.getnframes())
        p = pyaudio.PyAudio()
        while self.keep:
            stream = p.open(
                format=p.get_format_from_width(sampwidth),
                channels=channels,
                rate=framerate,
                output=True,
            )
            stream.write(frames)
            stream.stop_stream()
            stream.close()
        p.terminate()

    def start_singing(self) -> None:
        self.keep = True
        self.song_thread = threading.Thread(target=self.sing, daemon=True)
        self.song_thread.start()

    def stop_singing(self) -> None:
        self.keep = False
        self.song_thread.join()

    def message(self, text: str) -> None:
        self.prepare_root()
        self.start_singing()
        messagebox.showinfo("schedrem", text)
        self.stop_singing()
        self.root.destroy()

    def yesno(self, text: str) -> bool:
        self.prepare_root()
        ans = False
        self.start_singing()
        ans = messagebox.askyesno("schedrem", text)
        self.stop_singing()
        self.root.destroy()
        return ans

    def warning(self, text: str) -> None:
        self.prepare_root()
        messagebox.showwarning("schedrem", text)
        self.root.destroy()

    def error(self, text: str) -> None:
        self.prepare_root()
        messagebox.showerror("schedrem", text)
        self.root.destroy()


def error_message(errors: list[dict]) -> str:
    messages = []
    for error in errors:
        loc = ".".join([str(i) for i in error.get("loc", [])])
        msg = error.get("msg", "")
        ipt = error.get("input")
        if type(ipt) is dict and ipt.get("weekdaynames"):
            """SchedremConfig.validate_weekday() returns
            the entire SchedremConfig object, even though
            it only concerns TimeConfig.weekday and SchedremConfig.weekdaynames.
            Consider removing the unnecessary information.
            """
            ipt = ipt.get("weekdaynames")
        else:
            ipt = str(ipt)
        message = f"{loc}\n{msg}\ninput: {ipt}"
        messages.append(message)
    return "\n".join(messages)


def get_config_file() -> Path:
    r"""Search and return path to the config file.

    The files are searched in the order specified below and
    only the first one that is found is read.
    Defaults for Linux/macOS:
    "${XDG_CONFIG_HOME}/schedrem/config.yml"
    "${XDG_CONFIG_HOME}/schedrem/config.yaml"
    "${HOME}/.config/schedrem/config.yml"
    "${HOME}/.config/schedrem/config.yaml"
    Defaults for Windows:
    - "%USERPROFILE%\.config\schedrem\config.yml"
    - "%USERPROFILE%\.config\schedrem\config.yaml"
    - "%APPDATA%\schedrem\config.yml"
    - "%APPDATA%\schedrem\config.yaml"
    """
    if os.name == "posix":
        paths = (
            "${XDG_CONFIG_HOME}/schedrem/config.yml",
            "${XDG_CONFIG_HOME}/schedrem/config.yaml",
            "${HOME}/.config/schedrem/config.yml",
            "${HOME}/.config/schedrem/config.yaml",
        )
    elif os.name == "nt":
        paths = (
            "${USERPROFILE}/.config/schedrem/config.yml",
            "${USERPROFILE}/.config/schedrem/config.yaml",
            "${APPDATA}/schedrem/config.yml",
            "${APPDATA}/schedrem/config.yaml",
        )
    else:
        msg = "OS must be either POSIX or NT."
        raise OSError(msg)

    for p in paths:
        path = Path(os.path.expandvars(p))
        if path.exists():
            return path

    msg = f"Config file not found in {paths}."
    raise FileNotFoundError(msg)
